; Minesweeper for SEGA Dreamcast VM

; $Id: minesweeper.s,v 1.31 2003/07/19 13:18:17 sgust Exp $

;;NICE to have, but not yet implemented because of major changes:
;; 0 field recursion
;; bigger, scrolling playfield
;; sync highscores with other VMs

; Copyright (C) 2000 by Soeren Gust, sgust@ithh.informationstheater.de

; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.

; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

; You can get the latest version from http://soeren.informationstheater.de

; parts derived from tetris.s by Marcus Comstedt
	
	.include "sfr.i"

; Reset and interrupt vectors

; reset
	.org	0
	jmpf	start

; external irq 0
	.org	$3
	jmp	nop_irq

; external irq 1
	.org	$b
	jmp	nop_irq
	
; timer/counter 0 low 8 bit mode overflow
	.org	$13
	jmpf	timer0l_irq

; count clock
	.org	$1b
	jmp	clockint

; timer/counter 0 high / 16 bit mode overflow
	.org	$23
	jmp	timer0h_irq

; timer/counter 1 overflow
	.org	$2b
	jmpf	sound_irq

; SIO0
	.org	$33
	jmp	nop_irq

; SIO1
	.org	$3b
	jmp	nop_irq

; RFB
	.org	$43
	jmp	nop_irq

; port3 change: key pressed
	.org	$4b
	clr1	p3int,p3iflag		; clear flag
	clr1	p3int,p3ie		; disable further interrupts
	reti

nop_irq
	reti

; Firmware calls

	.org	$100
writeflash
	not1	ext,0
	jmpf	writeflash
	ret

	.org	$110
verifyflash
	not1	ext,0
	jmpf	verifyflash
	ret

	.org	$120
readflash
	not1	ext,0
	jmpf	readflash
	ret

; Clock irq: Count time, do housekeeping
	.org	$130
clockint:
	push	ie
	clr1	ie,mastie
	not1	ext,0
	jmpf	clockint
	pop	ie
	callf	myclockint
	reti
		
; Leave game
	.org	$1f0
goodbye:	
	not1	ext,0
	jmpf	goodbye

; Header
	.org	$200
; Name
	.byte	"VM  Minesweeper "
; Description
	.byte	"VM Minesweeper by Soeren Gust   "
; Creator application
	.byte	"aslc86k         "
; Icon header: number of frames
	.word	1
; Icon header: speed
	.word	0
; Icon header: eyecatch
	.word	0
; Icon header: crc, unused in game
	.word	0
; Length of file data
	.word	end,0
; Reserved
	.byte	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

	.include "icon.i"
	
; Main program

keyqstart	equ	$04	; first byte of keyboard queue
keyqend		equ	$0e	; last byte of keyboard queue
keyqin		equ	$0f	; pointer to insert into keyboard queue
keyqout		equ	$10	; pointer to read from keyboard queue
x		equ	$11	; x position for print
y		equ	$12	; y position for print
timelow		equ	$13	; low byte of gameclock
timehigh	equ	$14	; high byte of gameclock
timectl		equ	$15	; control byte for gameclock
mask1		equ	$16	; used by print
mask2		equ	$17	; used by print
char		equ	$18	; used by print
blinkflag	equ	$19	; counts and controls blinking in interrupts
blinkx		equ	$1a	; x position of blinking field in pixel
blinky		equ	$1b	; y position of blinking field in pixel
blinkchar	equ	$1c	; char to print for blinking field
curx		equ	$1d	; current x (0-5)
cury		equ	$1e	; current y (0-4)
screentoggle	equ	$1f	; which screen is displayed: title/highscore
unknownfields	equ	$20	; how many fields are visible
randlow		equ	$21	; bits 0-7 of random seed
randhigh	equ	$22	; bits 8-14 of random seed, highest bit unused
soundpointerl	equ	$23	; used by soundengine
soundpointerh	equ	$24	; used by soundengine
soundplaying	equ	$25	; used by soundengine
soundwait	equ	$26	; used by soundengine
bombs		equ	$27	; number of bombs hidden
keys		equ	$28	; key compare mask for keyboard irq
keyflags	equ	$29	; flags for A/B pressed on titlescreen
quitcount	equ	$2a	; counter for autoquit
lastkeyevent	equ	$2b	; store the latest key event for repeat
keyrepeat	equ	$2c	; counter/flag for key repeat
firststep	equ	$2d	; flag for first step

playfield	equ	$30	; - $4e: 30 byte playfield
; meaning of playfield:
; bit 7   flag
; bit 6   invisible
; bit 5   bomb
; bit 4-0 no of neighbor bombs

; constants
PF_LEFT		equ	0	; left screen position of playfield
PF_TOP		equ	0	; top screen position of playfield

; key numbers
KEY_UP		equ	1
KEY_DOWN	equ	2
KEY_LEFT	equ	3
KEY_RIGHT	equ	4
KEY_A		equ	5
KEY_B		equ	6
KEY_MODE	equ	7
KEY_SLEEP	equ	8
KEY_REL		equ	$80	; flag for key released

; bits in timectl
time_bit	equ	7
time_ovf	equ	2
time_disp	equ	1
time_run	equ	0

start
	clr1	ie,mastie	; disable all irq
	mov	#$a1,ocr	; speed control, subclock active, no div 2
	mov	#$09,mcr	; LCD on, no cursor, gfx mode
	mov	#$80,vccr	; LCD on, enable writing
	clr1	p3int,p3ie	; disable P3 irq
	clr1	p1,7		; ???
	mov	#$ff,p3		; enable pull-ups on P3 (buttons)

	; initialise random seed
	clr1	psw,rambk0	; system variable bank
	ld	$1d		; current time: seconds
	set1	psw,rambk0	; switch back to program variable bank
	st	randlow		; put seconds into low byte of random seed
	xor	#$ff		; invert it to make sure at least 2 bits in 
	st	randhigh	; the seed are different and put into high byte

	; init timer0 low interrupt frequency: 32768/6/32/128 = 1.34 per second
	mov	#0,t0con	; turn off everything
	mov	#$e0,t0prr	; prescaler to / 32
	mov	#$80,t0lr	; timer0 low to / 128

	; timer 0 high interrupt frequency: 32768/6/32/4 = 42.67 per second
	mov	#$fc,t0hr	; timer0 high to / 4

	; initialise keyboard
	mov	#$ff,keys		; no key pressed
	mov	#keyqstart,keyqin	; init queue pointer
	mov	#keyqstart,keyqout	; init queue pointer
	mov	#$80,lastkeyevent	; set latest key event to release
	mov	#$00,keyrepeat		; don't repeat
	set1	t0con,t0hie		; enable key interrupt
	set1	t0con,t0hrun		; start key timer

	; initialise rest
	mov	#5,bombs		; initial number of bombs
	mov	#0,timelow		; set clock
	mov	#0,timehigh		; to 0
	mov	#0,timectl		; clock stopped, no display, no overflow
	set1	ie,mastie		; enable all irq

title
	mov	#3,screentoggle		; first screen: title
	mov	#0,keyflags		; no keys pressed
	call	clearkey		; clear keys
	mov	#0,quitcount		; init auto power off counter
.screen
	; set clock for 10 seconds timeout
	clr1	timectl,time_run	; stop clock
	mov	#$90,timelow
	mov	#$09,timehigh
	clr1	timectl,time_bit	; lowest time bit
	clr1	timectl,time_ovf	; no overflow
	clr1	timectl,time_disp	; don't display clock
	set1	timectl,time_run	; start clock

	ld	screentoggle		; get screennumber
	bne	#3,.high1		; skip to highscore 1 if not title
	; display title screen
	mov	#<titlebitmap,trl
	mov	#>titlebitmap,trh
	call	copyscr
	jmp	.titleab		; continue with A+B checks

; first half of highscore table
.high1
	bne	#2,.high2		; skip to highscore 2 if not title
	clr1	ocr,subclk		; subclock off for faster display
	; display highscore 1 screen
	mov	#<showhighbitmap,trl
	mov	#>showhighbitmap,trh
	call	copyscr
	; fill in ranks
	mov	#2,acc			; start with 2 bombs
	mov	#0,x			; left to screen
	mov	#6,y			; start in second line
.high1ranks
	push	acc			; store character
	push	y			; store y
	call	print6x5		; print it
	pop	acc			; restore y
	add	#6			; next line
	st	y
	pop	acc			; restore character
	inc	acc			; next character
	bne	#6,.high1ranks		; display only till 5
	; first half of highscore data
	mov	#<highscores,trl
	mov	#>highscores,trh
	br	.showscore		; use routine of other half for contents

; second half of highscore table
.high2
	clr1	ocr,subclk		; subclock off for faster display
	; display highscore 2 screen
	mov	#<showhighbitmap,trl
	mov	#>showhighbitmap,trh
	call	copyscr
	; fill in ranks
	mov	#6,acc			; start with 2 bombs
	mov	#0,x			; left to screen
	mov	#6,y			; start in second line
.high2ranks
	push	acc			; store character
	push	y			; store y
	call	print6x5		; print it
	pop	acc			; restore y
	add	#6			; next line
	st	y
	pop	acc			; restore character
	inc	acc			; next character
	bne	#10,.high2ranks		; display only till 9
	; second half of highscore data
	mov	#<highscores+32,trl
	mov	#>highscores,trh

; show highscores starting at trh/trl
.showscore
	mov	#6,y
.scoreloop
	; print highest score nibble
	mov	#0,acc			; pointer offset to nibble
	ldc				; get data
	push	trl			; store registers
	push	trh
	push	y
	and	#$0f			; mask out nibble
	mov	#12,x
	call	print6x5		; print
	pop	y			; restore registers
	pop	trh
	pop	trl
	; print middle score nibble
	mov	#1,acc			; pointer offset
	ldc				; get data
	push	trl			; store registers
	push	trh
	push	y
	push	acc			; store lowest score byte
	ror				; rotate to exchange nibbles
	ror
	ror
	ror
	and	#$0f			; make out nibble
	mov	#18,x
	call	print6x5		; print it
	pop	acc			; restore lowest score byte
	pop	y			; restore registers partially
	; print lowest score nibble
	push	y			; store register
	and	#$0f
	mov	#24,x
	call	print6x5
	pop	y			; restore all registers
	pop	trh
	pop	trl
	; print name chars
	mov	#30,x
	mov	#2,acc
.scorename
	push	acc			; store pointer
	ldc				; get character
	push	trl			; store registers
	push	trh
	push	y
	call	print6x5
	pop	y			; restore registers
	pop	trh
	pop	trl
	ld	x			; get x position
	add	#6			; advance one column
	st	x
	pop	acc			; restore pointer
	inc	acc			; advance character
	bne	#5,.scorename		; until all 3 chars are printed
	; advance pointer and y
	ld	trl			; pointer low
	add	#8			; to next entry in table, overflow
	st	trl			; can't happen, block aligned
	ld	y			; y position
	add	#6			; to next line
	st	y
	bne	#30,.scoreloop		; until all 4 entries are printed
	set1	ocr,subclk		; drawing finished, back to subclock

; wait for A+B
.titleab
	call	randombit		; generate random bits while waiting
					; for better randomization
	call	getkey
	bne	#KEY_A,.nokeya		; test for A pressed
	bp	keyflags,0,.titleab	; A already set: loop
	set1	keyflags,0		; set flag for A
	ld	screentoggle
	be	#3,.nokeyb1
	mov	#3,screentoggle		; switch to titlescreen
	jmp	.screen			; restart timer and show screen
.nokeya
	bne	#KEY_A|KEY_REL,.nokeya1	; test for A released
	clr1	keyflags,0		; clear flag for A
	jmp	.screen			; restart timer and show screen
.nokeya1
	bne	#KEY_B,.nokeyb		; test for B pressed
	bp	keyflags,1,.titleab	; B already set: loop
	set1	keyflags,1		; set flag for B
	ld	screentoggle
	be	#3,.nokeyb1
	mov	#3,screentoggle		; switch to titlescreen
	jmp	.screen			; restart timer and show screen
.nokeyb
	bne	#KEY_B|key_rel,.nokeyb1	; test for B pressed
	clr1	keyflags,1		; clear flag for B
	jmp	.screen			; restart timer and show screen
.nokeyb1
	ld	keyflags		; get key state
	be	#3,setup		; A+B pressed
	bz	.nokeyab
	mov	#0,quitcount		; reset auto power off counter
.nokeyab
	bp	timectl,time_ovf,.time	; test for clock overflow
	br	.titleab

; timeout on titlescreen
.time
	ld	keyflags		; get A/B state
	bnz	.titleab		; A or B is pressed, dont change screen
	inc	quitcount		; increment power off counter
	ld	quitcount		; test power off counter
	be	#12,.poweroff		; nothing touched for 2 minutes
	dbnz	screentoggle,.screen1	; screentoggle counts 3,2,1,3,2,1
	mov	#3,screentoggle		; restart counter
.screen1
	jmp	.screen			; change screen

; quit game to conserve batteries
.poweroff
	jmp	goodbye

; "select number of bombs"
setup
	call	clearkey		; clear key queue
	mov	#<bombbitmap,trl	; display the
	mov	#>bombbitmap,trh	; select # of bombs
	call	copyscr			; screen
	; set clock for 30 seconds timeout
	clr1	timectl,time_run	; stop clock
	mov	#$70,timelow
	mov	#$09,timehigh
	clr1	timectl,time_bit	; lowest time bit
	clr1	timectl,time_ovf	; no overflow
	clr1	timectl,time_disp	; don't display clock
	set1	timectl,time_run	; start clock
.setbombdraw
	; draw number of bombs currently selected
	mov	#43,x
	mov	#20,y
	ld	bombs
	call	print6x5
.setbombloop
	bn	timectl,time_ovf,.setbombtime	; check for timeout
	jmp	title				; back to title if timeout
.setbombtime
	; handle key presses
	call	getkey				; get key
	bne	#KEY_b,.setbombcont		; B pressed?
	jmp	title				; back to title if B
.setbombcont
	bne	#KEY_LEFT,.setbombnoleft	; skip if left not pressed
	; left pressed
	ld	bombs
	be	#2,.setbombskip1		; minimum 2 bombs
	dec	bombs
.setbombskip1
	br	.setbombdraw			; draw new bomb count
.setbombnoleft
	bne	#KEY_RIGHT,.setbombnoright	; skip if right not pressed
	; right pressed
	ld	bombs
	be	#9,.setbombskip2		; maximum 9 bombs
	inc	bombs
.setbombskip2
	br	.setbombdraw			; draw new bomb count
.setbombnoright
	bne	#KEY_A,.setbombloop		; if A not pressed wait

; A is pressed, prepare to play
	mov	#1,firststep		; its the first step
	mov	#30,acc			; number of fields
	sub	bombs			; minus number of bombs
	st	unknownfields		; = number of fields to explore
	call	clearfield		; init playfield
	ld	bombs
	call	hidebombs		; hide the bombs

	; init for playloop
	call	clrscr			; clear screen
	call	printfield		; and display field

	; display number of hidden bombs
	mov	#39,x			; x position bomb symbol
	mov	#12,y			; y position bomb symbol
	mov	#11,acc			; bomb symbol
	call	print6x5		; print the bomb symbol
	mov	#39,x			; x position number
	mov	#18,y			; y position number
	ld	bombs			; get number of bombs
	call	print6x5		; print number of bombs

	mov	#0,curx			; left x pos
	mov	#0,cury			; top y pos

	mov	#PF_LEFT,blinkx		; left screen pos
	mov	#PF_TOP,blinky		; top screen pos
	ld	playfield		; first field
	call	decodefield		; convert to char
	st	blinkchar		; store for blinking
	; draw cursor on first field
	ld	blinkx
	st	x
	ld	blinky
	st	y
	mov	#13,acc
	call	print6x5

	; start cursor blink
	mov	#1,blinkflag		; field
	set1	t0con,t0lie		; enable interrupts
	set1	t0con,t0lrun		; start timer0 low

	; init my clock
	mov	#0,timectl
	mov	#0,timelow
	mov	#0,timehigh
	; display 000 as clock
	mov	#37,x
	mov	#0,y
	mov	#0,acc
	call	print3x5
	mov	#41,x
	mov	#0,y
	mov	#0,acc
	call	print3x5
	mov	#45,x
	mov	#0,y
	mov	#0,acc
	call	print3x5
	; start clock
	clr1	timectl,time_bit	; lowest time bit
	clr1	timectl,time_ovf	; no overflow
	set1	timectl,time_disp	; display clock
	set1	timectl,time_run	; start clock
	call	clearkey		; clear key queue

playloop
	; test keys
	call	getkey			; get key
	be	#KEY_UP,playup		; pressed up
	be	#KEY_DOWN,playdown	; pressed down
	be	#KEY_LEFT,playleft	; pressed left
	be	#KEY_RIGHT,playright	; pressed right
	be	#KEY_A,playa		; pressed A
	be	#KEY_B,playb		; pressed B

	; test for time over
	bn	timectl,time_ovf,playloop	; again
	jmp	explode			; timer expired, you lose

; move cursor up
playup
	ld	cury			; y position
	be	#0,playloop		; top border?
	dec	cury			; move up
	jmp	newpos			; set new position

; move cursor down
playdown
	ld	cury			; y position
	be	#4,playloop		; bottom border?
	inc	cury			; move down
	jmp	newpos			; set new position

; move cursor left
playleft
	ld	curx			; x position
	be	#0,playloop		; left border?
	dec	curx			; move left
	jmp	newpos			; set new position

; move cursor right
playright
	ld	curx			; x position
	be	#5,playloop		; right border?
	inc	curx			; move right
	jmp	newpos			; set new position

; button A pressed: show field
playa
	call	getplaypos		; get position
	st	0			; used as pointer to ram
	ld	@r0
	bp	acc,7,.playaend		; flag is set, don't touch
	bn	acc,6,.playaend		; already visible, ignore
	bn	acc,5,.playa1		; skip if no bomb
	bp	firststep,0,.shiftfirst	; check for first step
	jmp	explode			; explosion, game over
.shiftfirst
	; we hit a bomb on first step and have to move the bomb
	mov	#1,acc
	push	0			; store pointer
	call	hidebombs		; hide 1 more bomb
	pop	0
	mov	#0,@r0			; clear current bomb
.playa1
	; show current field
	mov	#0,firststep		; first step is done
	call	countbombs		; count the bombs around
	st	@r0			; store in field
	call	decodefield		; to char
	clr1	t0con,t0lie		; stop blink interrupt
	st	blinkchar		; set new char
	ld	blinkx			; x position of field ...
	st	x			; ... ready for print
	ld	blinky			; y position of field ...
	st	y			; ... ready for print
	ld	blinkchar		; char
	call	print6x5		; print it
	; restart the blinking at field just drawn phase
	clr1	t0con,t0lrun		; stop timer
	clr1	t0con,t0lovf		; ignore past interrupts
	mov	#0,blinkflag		; field
	set1	t0con,t0lie		; reenable blink interrupt
	set1	t0con,t0lrun		; start timer (gets reloaded)
	; now count free field
	dec	unknownfields		; we know one more
	ld	unknownfields
	bnz	.playaend		; still unknown fields?
	jmp	gamewon			; all non bomb field are visible
.playaend
	jmp	playloop

; button B pressed: set/clear flag
playb
	call	getplaypos		; get position
	st	0			; used as pointer to ram
	ld	@r0			; get field
	bn	acc,6,.playbend		; already visible
	xor	#$80			; change flag display
	st	@r0			; store field
	call	decodefield		; to char
	clr1	t0con,t0lie		; stop blink interrupt
	st	blinkchar		; set new char
	ld	blinkx			; x position of field ...
	st	x			; ... ready for print
	ld	blinky			; y position of field ...
	st	y			; ... ready for print
	ld	blinkchar		; char
	call	print6x5		; print it
	; restart the blinking at field just drawn phase
	clr1	t0con,t0lrun		; stop timer
	clr1	t0con,t0lovf		; ignore past interrupts
	mov	#0,blinkflag		; field
	set1	t0con,t0lie		; reenable blink interrupt
	set1	t0con,t0lrun		; start timer (gets reloaded)
.playbend
	jmp	playloop		; continue gameplay

; calculate new position, move cursor etc.
newpos
	clr1	t0con,t0lie		; disable blink irq
	; draw old field
	ld	blinkx			; x position of field ...
	st	x			; ... ready for print
	ld	blinky			; y position of field ...
	st	y			; ... ready for print
	ld	blinkchar
	call	print6x5

	; calculate new blinkx, blinky
	ld	curx			; char position
	rol				; *2
	rol				; *4
	add	curx			; *5
	add	curx			; *6
	add	#PF_LEFT		; *6+offset
	st	blinkx			; = pixel position
	ld	cury			; char position
	rol				; *2
	rol				; *4
	add	cury			; *5
	add	cury			; *6
	add	#PF_TOP			; *6+offset
	st	blinky			; = pixel position
	; get new t0field
	call	getplaypos
	st	0
	ld	@r0			; get field
	call	decodefield		; decode it
	st	blinkchar		; new char for blink

	; draw cursor on new field
	ld	blinkx
	st	x
	ld	blinky
	st	y
	mov	#13,acc
	call	print6x5

	; restart the blinking at cursor just drawn phase
	clr1	t0con,t0lrun		; stop timer
	clr1	t0con,t0lovf		; ignore past interrupts
	mov	#1,blinkflag		; cursor
	set1	t0con,t0lie		; reenable blink interrupt
	set1	t0con,t0lrun		; start timer (gets reloaded)

	jmp	playloop

; all fields are free, game won
gamewon
	clr1	t0con,t0lie		; disable blink irq
	clr1	t0con,t0lrun		; stop blink timer
	clr1	timectl,time_run	; stop clock
	clr1	timectl,time_disp	; stop clock display

	; test highscore
	mov	#<highscores,trl	; pointer to highscore table
	mov	#>highscores,trh
	ld	bombs			; number of bombs
	sub	#2			; -2 = entry number
	rol				; *2
	rol				; *4
	rol				; *8 = offset to entry
	st	b
	; now calculate highscore-score-1
	inc	acc			; low byte of highscore
	ldc
	set1	psw,cy
	subc	timelow
	ld	b
	ldc
	subc	timehigh
	; check for overflow from highest nibble
	and	#$f0
	bz	newhigh			; no overflow, score < highscore

	clr1	ocr,subclk		; turn off subclock for message

	; show winning message
	mov	#36,x
	mov	#12,y
	mov	#14,acc
	call	print6x5
	mov	#42,x
	mov	#12,y
	mov	#15,acc
	call	print6x5
	mov	#36,x
	mov	#18,y
	mov	#16,acc
	call	print6x5
	mov	#42,x
	mov	#18,y
	mov	#17,acc
	call	print6x5

	set1	ocr,subclk		; back to subclock

	; play the winning sound
	mov	#<wintune,trl
	mov	#>wintune,trh
	call	playtune

	jmp	gameend

; got a new highscore
newhigh
	push	b			; offset to entry in highscore table
	; read flash rom segment into into $80-$ff
	push	ie			; save registers
	push	ocr
	clr1	ie,mastie		; disable interrupts
	clr1	ocr,subclk		; subclock off, necessary!!
	mov	#0,$7d
	mov	#>highscores,$7e
	mov	#<highscores,$7f
	call	readflash
	pop	ocr
	pop	ie
	; display new highscore screen
	mov	#<newhighbitmap,trl
	mov	#>newhighbitmap,trh
	call	copyscr
	; display score value
	mov	#16,x
	mov	#13,y
	ld	timehigh
	and	#$f
	call	print6x5
	mov	#22,x
	mov	#13,y
	ld	timelow
	rol
	rol
	rol
	rol
	and	#$f
	call	print6x5
	mov	#28,x
	mov	#13,y
	ld	timelow
	and	#$f
	call	print6x5
	; fill in score
	pop	b			; pointer offset to entry
	ld	b
	add	#$80
	st	0			; pointer to entry in ram
	ld	timehigh
	st	@r0
	inc	0
	ld	timelow
	st	@r0
	inc	0			; points now to first name char in ram
	; set name to AAA
	push	0			; store pointer register
	mov	#23,acc			; A
	st	@r0			; first of name
	inc	0
	st	@r0			; second of name
	inc	0
	st	@r0			; third of name
	pop	0			; restore pointer: back to first
	; set position
	mov	#16,x
	mov	#27,y

	; play the highscore sound
	mov	#<hightune,trl
	mov	#>hightune,trh
	call	playtune
	; wait until sound stops
.soundend
	bp	soundplaying,0,.soundend
	call	clearkey		; clear key queue
	set1	keyrepeat,7		; enable keyboard repeat

.newhighprint
	push	y
	ld	@r0
	call	print6x5
	pop	y
	; loop to enter name into highscore
.newhighloop
	call	getkey
	be	#KEY_A,.newhighforw	; A next char
	be	#KEY_B,.newhighback	; B prev char
	be	#KEY_DOWN,.newhighprev	; previous character
	be	#KEY_UP,.newhighnext	; next character
	be	#KEY_LEFT,.newhighprev	; previous character
	be	#KEY_RIGHT,.newhighnext	; next character
	br	.newhighloop

; next char from alphabet
.newhighnext
	ld	@r0
	bne	#49,.newhighnext1	; !=.?
	mov	#22,@r0			; A after inc
.newhighnext1
	inc	@r0
	jmp	.newhighprint

; previous char from alphabet
.newhighprev
	ld	@r0
	bne	#23,.newhighprev1	; !=A?
	mov	#50,@r0
.newhighprev1
	dec	@r0
	jmp	.newhighprint

; back one position in name
.newhighback
	; clear current char
	push	y
	mov	#22,acc			; SPACE
	call	print6x5
	pop	y
	; move x
	ld	x
	sub	#6
	st	x
	; advance pointer
	dec	0
	ld	0
	and	#7
	bne	#1,.newhighprint	; test for legal position
	; oops, one position too far left, undo changes
	inc	0
	ld	x
	add	#6
	st	x
	br	.newhighprint		; print char again

; forward one position in name
.newhighforw
	; advance pointer
	inc	0
	ld	0
	and	#7
	be	#5,.newhighend		; test for last in name
	; move x
	ld	x
	add	#6
	st	x
	jmp	.newhighprint
.newhighend
	clr1	keyrepeat,7		; disable key repeat
	; write flash rom segment back
	mov	#1,$7c			; wait to stabilize
	mov	#0,$7d			; highest nibble: game bank
	mov	#>highscores,$7e
	mov	#<highscores,$7f
	push	ie			; save registers
	push	ocr
	clr1	ie,mastie		; disable interrupts
	clr1	ocr,subclk		; subclock off, necessary!!
	mov	#0,acc			; softvms <= 1.8 only
	call	writeflash		; write to flash
	pop	ocr			; restore registers
	pop	ie
	bnz	.flasherror		; test for error
	push	ie			; save registers
	push	ocr
	clr1	ie,mastie		; disable interrupts
	clr1	ocr,subclk		; subclock off, necessary!!
	call	verifyflash		; verify
	pop	ocr			; restore registers
	pop	ie
	bnz	.flasherror		; test for error
	jmp	title

; something went wrong while flashing
.flasherror
	; show screen
	mov	#<flasherrbitmap,trl
	mov	#>flasherrbitmap,trh
	call	copyscr
.flasherrwait
	; wait for A
	call	getkey
	bne	#KEY_A,.flasherrwait
	jmp	title

; hit a mine, game over
explode
	clr1	t0con,t0lie		; disable blink irq
	clr1	t0con,t0lrun		; stop blink timer
	clr1	timectl,time_run	; stop clock
	clr1	timectl,time_disp	; stop clock display

	; show playfield, clear flags, count bombs on every field
	clr1	ocr,subclk		; subclock off: faster display refresh
	mov	#playfield,0		; first field
	mov	#PF_TOP,y		; top y position
	mov	#0,cury			; counter for lines
.explodey
	mov	#PF_LEFT,x		; left x position
	mov	#0,curx			; counter for columns
.explodex
	ld	@r0			; get contents
	clr1	acc,7			; clear flag (in case of bomb)
	clr1	acc,6			; clear invisibility (in case of bomb)
	bp	acc,5,.explode		; skip counting if bomb
	call	countbombs		; count the bombs
.explode
	st	@r0			; store in field (not really necessary)
	call	decodefield		; convert to char
	push	y
	call	print6x5		; print it
	pop	y

	inc	0			; next field
	ld	x			; advance ...
	add	#6			; ... x position ...
	st	x			; ... to next field

	inc	curx			; x counter
	ld	curx			; get x counter
	bne	#6,.explodex		; are all 6 columns printed?

	ld	y			; advance ...
	add	#6			; ... y position ...
	st	y			; ... to next line
	inc	cury			; y counter
	ld	cury			; get y counter
	bne	#5,.explodey		; are all 5 lines printed?

	; delete no of bombs
	mov	#39,x
	mov	#12,y
	mov	#22,acc
	call	print6x5
	mov	#39,x
	mov	#18,y
	mov	#22,acc
	call	print6x5

	; show game over message
	; "GAME" is normal font
	mov	#37,x
	mov	#6,y
	mov	#29,acc
	call	print6x5
	mov	#37,x
	mov	#12,y
	mov	#23,acc
	call	print6x5
	mov	#37,x
	mov	#18,y
	mov	#35,acc
	call	print6x5
	mov	#37,x
	mov	#24,y
	mov	#27,acc
	call	print6x5
	; "OVER" is shifted one pixel, so special font
	mov	#42,x
	mov	#6,y
	mov	#18,acc
	call	print6x5
	mov	#42,x
	mov	#12,y
	mov	#19,acc
	call	print6x5
	mov	#42,x
	mov	#18,y
	mov	#20,acc
	call	print6x5
	mov	#42,x
	mov	#24,y
	mov	#21,acc
	call	print6x5

	set1	ocr,subclk		; back to subclock

	; play the explode sound
	mov	#<explodetune,trl
	mov	#>explodetune,trh
	call	playtune

; here starts the preparation for a new game for all game endings
gameend
	call	clearkey		 ; clear keyboard queue
	; set clock to wait for 20 seconds
	clr1	timectl,time_run	; stop clock
	mov	#$80,timelow
	mov	#$09,timehigh
	clr1	timectl,time_bit	; lowest time bit
	clr1	timectl,time_ovf	; no overflow
	clr1	timectl,time_disp	; don't display clock
	set1	timectl,time_run	; start clock

	; wait until sound stops
.soundend
	bp	soundplaying,0,.soundend

.wait
	bp	timectl,time_ovf,.title	; if timer overflow go to title
	call	getkey			; get keys
	be	#KEY_A,.setup		; if A pressed go to select bombs
	be	#KEY_B,.title		; if B pressed go to title
	br	.wait			; loop
.setup
	jmp	setup
.title
	jmp	title

; --- playfield routines starts here ---

; clear the playfield: set all to $40
; destroys: acc, 0, b
clearfield
	mov	#$40,acc	; value to fill field
	mov	#30,b		; number of fields to fill
	mov	#playfield,0	; first field to fill
.clearfield
	st	@r0		; set field
	inc	0		; advance pointer
	dbnz	b,.clearfield	; next field

	ret			; finished

; hides bombs in playfield
; meaning of playfield:
; bit 7   flag
; bit 6   invisible
; bit 5   bomb
; bit 4-0 no of neighbor bombs
; get number of bombs to hide in acc
; destroys: acc, 0, b
hidebombs
	push	acc			; store number of bombs
	; generate random field number by getting 5 bits and rejecting 30/31
.makefield
	mov	#0,acc			; clear acc, will contain fieldnumber
	mov	#5,b			; 5 bits
.getrand
	call	randombit		; get one bit
	rolc				; put together
	dbnz	b,.getrand		; until 5 bits are generated
	be	#30,.makefield		; field 30 does not exist
	be	#31,.makefield		; field 31 does not exist
	add	#playfield		; add playfield offset
	st	0			; store as pointer
	ld	@r0			; get field contents
	bp	acc,5,.makefield	; if bomb try again
	mov	#$60,acc		; bomb, invisible
	st	@r0			; put it into field
	pop	acc			; get number of bomb from stack
	dec	acc			; one more hidden
	bz	.hidebombsend		; all hidden?
	push	acc			; store number of bombs left to hide
	br	.makefield		; generate next fieldnumber
.hidebombsend
	ret				; finished

; counts the bombs around @r0, position (curx, cury)
; destroys b
countbombs
	mov	#0,b			; bomb count
	; left, top
	ld	curx
	be	#0,.count3
	ld	cury
	be	#0,.count1
	push	0
	ld	0
	sub	#7
	st	0
	ld	@r0
	pop	0
	bn	acc,5,.count1
	inc	b			; bomb found
.count1
	; left: no check, because already skipped to count3
	push	0
	dec	0
	ld	@r0
	pop	0
	bn	acc,5,.count2
	inc	b			; bomb found
.count2
	; left, down: no check for left, because already skipped to count3
	ld	cury
	be	#4,.count3
	push	0
	ld	0
	add	#5
	st	0
	ld	@r0
	pop	0
	bn	acc,5,.count3
	inc	b			; bomb found
.count3
	; top
	ld	cury
	be	#0,.count4
	push	0
	ld	0
	sub	#6
	st	0
	ld	@r0
	pop	0
	bn	acc,5,.count4
	inc	b			; bomb found
.count4
	; down
	ld	cury
	be	#4,.count5
	push	0
	ld	0
	add	#6
	st	0
	ld	@r0
	pop	0
	bn	acc,5,.count5
	inc	b			; bomb found
.count5
	; right, top
	ld	curx
	be	#5,.count8
	ld	cury
	be	#0,.count6
	push	0
	ld	0
	sub	#5
	st	0
	ld	@r0
	pop	0
	bn	acc,5,.count6
	inc	b			; bomb found
.count6
	; right: no check needed
	push	0
	inc	0
	ld	@r0
	pop	0
	bn	acc,5,.count7
	inc	b			; bomb found
.count7
	; right, down: no check for right
	ld	cury
	be	#4,.count8
	push	0
	ld	0
	add	#7
	st	0
	ld	@r0
	pop	0
	bn	acc,5,.count8
	inc	b			; bomb found
.count8
	ld	b			; get bomb count
	ret

; calculate playfield position from curx, cury
getplaypos
	ld	cury		; y
	rol			; y*2
	rol			; y*4
	add	cury		; y*5
	add	cury		; y*6
	add	curx		; y*6+x
	add	#playfield	; playfield+y*6+x = pointer to current playfield
	ret

; decode fieldcontents in acc to char in acc
; meaning of playfield:
; bit 7   flag
; bit 6   invisible
; bit 5   bomb
; bit 4-0 no of neighbor bombs
decodefield
	bp	acc,7,.decodeflag
	bp	acc,6,.decodeinvisible
	bp	acc,5,.decodebomb
	ret
.decodeflag
	mov	#12,acc
	ret
.decodeinvisible
	mov	#10,acc
	ret
.decodebomb
	mov	#11,acc
	ret

; --- playfield routines end here ---

; --- drawing routines starts here ---

; Clear the screen
; code from TETRIS with small changes
; destroys acc,xbnk
clrscr:	
	push	ocr		; store oscillator
	clr1	ocr,subclk	; we access screen, no more subclock
	push	2		; used as counter
	mov	#0,xbnk		; LCD bank 0
.cbank:	mov	#$80,2		; init counter
.cloop:	mov	#0,@R2		; clear byte in LCD ram
	inc	2		; advance counter
	ld	2		; get counter
	and	#$f		; test for end of even row
	bne	#$c,.cskip	; no, continue
	ld	2		; end of even row, get counter
	add	#4		; start of next row
	st	2		; put counter
.cskip:	ld	2		; get counter
	bnz	.cloop		; test for end of LCD bank
	bp	xbnk,0,.cexit	; skip to end if last LCD bank
	mov	#1,xbnk		; continue in LCD bank 1
	br	.cbank		; start again
.cexit:	pop	2		; restore R2
	pop	ocr		; restore oscillator
	ret

; copy a bitmap to the screen
; trl/trh pointer to bitmap
; destroys acc, b, c, 2, trl/trh, xbnk
copyscr
	push	ocr		; store registers
	clr1	ocr,subclk	; we access screen, no more subclock
	mov	#0,b		; rom pointer offset
	mov	#0,xbnk		; start in lcd bank 0
	mov	#$80,2		; init pointer to lcd
.cpyloop
	mov	#12,c		; 12 bytes = 2 lines = 1 double-line
.cpyline
	ld	b		; get rom-offset
	ldc			; get byte from rom
	st	@r2		; put into screen
	inc	2		; advance lcd pointer
	inc	b		; advance rom-offset
	dbnz	c,.cpyline	; next byte

	ld	2		; pointer to lcd
	add	#4		; advance to start of next double-line
	st	2		; new pointer to lcd
	bn	psw,cy,.cpyloop	; next line
	mov	#$80,2		; reinit pointer to lcd
	inc	xbnk		; next lcd bank
	bn	xbnk,1,.cpyloop	; continue copying if in lcd bank 0/1

	pop	ocr
	ret

; print a 6x5 pixel char
; x/y position, acc char
; destroys y, xbnk, acc, trl/trh
print6x5
	push	ocr		; save registers
	clr1	ocr,subclk	; we access screen, no more subclock
	push	2
	push	b
	push	c

	; calculate charpointer offset
	st	c
	rol			; *2
	rol			; *4
	add	c		; *5
	add	#<font5x5
	st	trl
	mov	#0,acc
	addc	#>font5x5
	st	trh
	mov	#0,char
	
	; calculate LCD line
	ld	y
	rol
	rol
	rol			; y * 8
	and	#$70		; mask out bits 3-1
	add	#$80		; offset to access LCD
	bn	y,0,.print_line	; test for even line (use bit 0)
	add	#6		; add 6 for even
.print_line
	st	2		; LCD pointer to begin of line
	; add x position
	ld	x		; get x position in bits
	ror
	ror
	ror			; /8
	and	#$7		; x position in bytes
	add	2
	st	2		; LCD pointer to begin of char

	; calculate LCD bank
	ld	y
	ror
	ror
	ror
	ror
	and	#1		; use only bit 4 of y
	st	xbnk		; as LCD bank

	; bitmasks
	mov	#$fc,mask1	; 6 highest bits set
	mov	#$00,mask2
	ld	x
	and	#7
	st	c
	bz	.print_skiprot	; if already aligned skip rotation
	clr1	psw,cy		; clear carry
.print_bitmask2
	; rotate bitmasks right from mask1 to mask2
	ld	mask1		; get mask for 1st byte
	rorc			; shift right
	st	mask1		; store it again
	ld	mask2		; get mask for 2nd byte
	rorc			; shift lowest byte from 1st into highest from 2nd
	st	mask2		; store it
	dbnz	c,.print_bitmask2

.print_skiprot
	mov	#5,b		; height
.print_loop
	ld	char		; get pointer to font
	ldc			; get char
	push	acc		; store char
	ld	x		; get x position
	and	#7		; number of shifts
	bz	.print_noshift	; skip shifting if aligned
	st	c		; put into c
	pop	acc		; restore char
.print_shift
	ror			; rotate
	dbnz	c,.print_shift	; c times
	br	.print_print
.print_noshift
	pop	acc		; restore unshifted char data
.print_print
	; handle first byte
	push	acc		; shifted font: needed later again
	and	mask1		; mask out
	st	c		; store shifted and masked font in c
	ld	mask1		; get mask
	xor	#$ff		; invert
	and	@R2		; mask byte from LCD
	or	c		; insert char
	st	@R2		; put into LCD
	; handle 2nd byte
	inc	2		; next LCD byte
	pop	c		; pop shifted font from stack
	ld	mask2		; get mask
	bz	.print_no2nd	; skip 2nd byte completely if possible
	and	c		; mask font
	st	c		; store font
	ld	mask2		; get mask
	xor	#$ff		; invert mask
	and	@R2		; mask byte from LCD
	or	c		; insert char
	st	@R2		; put into LCD
.print_no2nd
	; advance to next LCD line
	ld	2		; get LCD position
	add	#5		; next line
	bn	y,0,.print_lineskip
	add	#4		; current row was even
.print_lineskip
	bp	acc,7,.print_nobank	; still in same bank
	inc	xbnk		; next bank
	add	#$80		; correct position for new bank
.print_nobank
	st	2		; store new LCD position
	inc	char		; advance pointer to font
	inc	y		; position tracking
	dbnz	b,.print_loop

	pop	c
	pop	b
	pop	2
	pop	ocr
	ret

; print a 3x5 pixel char, can't cross byte boundaries!!
; x/y position, acc char
; destroys y, xbnk, acc, trl/trh
print3x5
	push	ocr		; save registers
	clr1	ocr,subclk	; we access screen, no more subclock
	push	2
	push	b
	push	c

	; calculate charpointer offset
	st	c
	rol			; *2
	rol			; *4
	add	c		; *5
	add	#<font3x5
	st	trl
	mov	#0,acc
	addc	#>font3x5
	st	trh
	mov	#0,char
	
	; calculate LCD line
	ld	y
	rol
	rol
	rol			; y * 8
	and	#$70		; mask out bits 3-1
	add	#$80		; offset to access LCD
	bn	y,0,.print_line	; test for even line (use bit 0)
	add	#6		; add 6 for even
.print_line
	st	2		; LCD pointer to begin of line
	; add x position
	ld	x		; get x position in bits
	ror
	ror
	ror			; /8
	and	#$7		; x position in bytes
	add	2
	st	2		; LCD pointer to begin of char

	; calculate LCD bank
	ld	y
	ror
	ror
	ror
	ror
	and	#1		; use only bit 4 of y
	st	xbnk		; as LCD bank

	; bitmasks
	mov	#$e0,mask1	; 3 highest bits set
	ld	x
	and	#7
	st	c
	bz	.print_skiprot	; if already aligned skip rotation
	ld	mask1		; get mask for 1st byte
.print_bitmask2
	ror			; shift right
	dbnz	c,.print_bitmask2
	st	mask1		; store it again

.print_skiprot
	mov	#5,b		; height
.print_loop
	ld	char		; get pointer to font
	ldc			; get char
	push	acc		; store char
	ld	x		; get x position
	and	#7		; number of shifts
	bz	.print_noshift	; skip shifting if aligned
	st	c		; put into c
	pop	acc		; restore char
.print_shift
	ror			; rotate
	dbnz	c,.print_shift	; c times
	br	.print_print
.print_noshift
	pop	acc		; restore unshifted char data
.print_print
	; handle only byte
	and	mask1		; mask out
	st	c		; store shifted and masked font in c
	ld	mask1		; get mask
	xor	#$ff		; invert
	and	@R2		; mask byte from LCD
	or	c		; insert char
	st	@R2		; put into LCD
.print_no2nd
	; advance to next LCD line
	ld	2		; get LCD position
	add	#6		; next line
	bn	y,0,.print_lineskip
	add	#4		; current row was even
.print_lineskip
	bp	acc,7,.print_nobank	; still in same bank
	inc	xbnk		; next bank
	add	#$80		; correct position for new bank
.print_nobank
	st	2		; store new LCD position
	inc	char		; advance pointer to font
	inc	y		; position tracking
	dbnz	b,.print_loop

	pop	c
	pop	b
	pop	2
	pop	ocr
	ret

; print the playfield
; destroys x,y, 0
printfield
	; store registers
	push	ocr
	clr1	ocr,subclk	; subclock off, for faster display refresh
	push	b
	push	c

	mov	#playfield,0	; first field
	mov	#0,y		; top y position
	mov	#0,c		; counter for lines
.printfy
	mov	#0,x		; left x position
	mov	#0,b		; counter for columns
.printfx
	ld	@r0		; get contents
	call	decodefield	; decode
	push	y
	call	print6x5	; print it
	pop	y

	inc	0		; next field
	ld	x		; advance ...
	add	#6		; ... x position ...
	st	x		; ... to next field

	inc	b		; x counter
	ld	b		; get x counter
	bne	#6,.printfx	; are all 6 columns printed?

	ld	y		; advance ...
	add	#6		; ... y position ...
	st	y		; ... to next line
	inc	c		; y counter
	ld	c		; get y counter
	bne	#5,.printfy	; are all 5 lines printed?

	; restore regs
	pop	c
	pop	b
	pop	ocr
	ret

; --- drawing routines end here ---

; --- misc routines start here ---

; timer0 low interrupt: manage the blinking cursor
timer0l_irq
	bn	t0con,t0lovf,.timer0ret	; skip if not timer0 low ovl
	clr1	t0con,t0lovf		; clear interrupt reason
	push	psw			; store all used regs on stack
	push	acc
	push	trl
	push	trh
	push	x
	push	y
	push	char
	push	mask1
	push	mask2
	push	xbnk

	set1	psw,rambk0		; game ram bank
	ld	blinkx			; x position of field ...
	st	x			; ... ready for print
	ld	blinky			; y position of field ...
	st	y			; ... ready for print

	not1	blinkflag,0		; invert blink flag
	ld	blinkchar		; get field char
	bn	blinkflag,0,.timer0l_print	; draw field if flag=1

	mov	#13,acc			; cursor char
.timer0l_print
	call	print6x5		; print it

	clr1	t0con,t0lrun		; stop timer and restart timer,
	set1	t0con,t0lrun		; it ran too fast in print6x5 (RC clk)

	pop	xbnk			; restore used regs
	pop	mask2
	pop	mask1
	pop	char
	pop	y
	pop	x
	pop	trh
	pop	trl
	pop	acc
	pop	psw
.timer0ret
	reti

; my own time keeping routine, hooked into system time counter
myclockint
	push	acc				; store registers
	push	psw
	bn	timectl,time_run,.myclockend	; if clock should not run, skip
	not1	timectl,time_bit		; lowest bit, twice per second
	bp	timectl,time_bit,.myclockend	; skip if first call
	inc	timelow				; count time byte low
	ld	timelow
	and	#$0f				; lowest time nibble
	bne	#$0a,.myclockprint		; test for BCD overflow
	mov	#$06,acc			; lowest nibble overflow
	add	timelow
	st	timelow				; correct low byte
	and	#$f0				; middle time nibble
	bne	#$a0,.myclockprint		; test for BCD overflow
	mov	#0,timelow			; overflow, clear low byte
	inc	timehigh			; count time byte high
	ld	timehigh
	and	#$0f				; highest time nibble
	bne	#$0a,.myclockprint		; check for BCD overflow
	; BCD overflow, no more nibbles left, stop and set flag
	mov	#0,timelow
	mov	#0,timehigh
	set1	timectl,time_ovf		; overflow
	clr1	timectl,time_disp		; stop display, last = 999
	clr1	timectl,time_run		; stop clock
.myclockprint
	bn	timectl,time_disp,.myclockend	; if clock should display, skip
	; print time in BCD
	push	x				; store more registers
	push	y
	push	mask1
	push	char
	push	xbnk
	push	trl
	push	trh
	mov	#37,x
	mov	#0,y
	ld	timehigh
	and	#$0f
	call	print3x5			; print highest nibble
	mov	#41,x
	mov	#0,y
	ld	timelow
	ror
	ror
	ror
	ror
	and	#$0f
	call	print3x5			; print middle nibble
	mov	#45,x
	mov	#0,y
	ld	timelow
	and	#$0f
	call	print3x5			; print lowest nibble
	pop	trh				; restore regs
	pop	trl
	pop	xbnk
	pop	char
	pop	mask1
	pop	y
	pop	x
.myclockend
	pop	psw				; restore more regs
	pop	acc
	ret

; generate 1 random bit, return in carry
; uses the primitiv polynom 1 + V^14 + V^15 in GF(2)
; the 1 can be ignored if we set U=1 constant
; needs 15 bit shift register in randhigh, randlow
; repeats after 2^15 - 1 bits
randombit
	push	acc		; store register
	ld	randhigh
	rol			; acc:6 is now V^14
	xor	randhigh	; acc:6 is now V^14 + V^15 = new bit
	rolc			; new bit in acc:7
	rolc			; new bit now in Carry
	push	psw		; store Carry, its our return value
	ld	randlow
	rolc			; shift new bit into seed
	st	randlow
	ld	randhigh
	rolc			; shift rest of seed
	st	randhigh
	pop	psw		; get out return value, bit in Carry
	pop	acc		; restore register
	ret

; --- misc routines end here ---

; --- keyboard code starts here ---

; handle periodic keyboard interrupt
timer0h_irq
	bp	t0con,t0hovf,.timer0do	; skip if timer0 high ovl
	reti
.timer0do
	clr1	t0con,t0hovf		; clear interrupt reason
	push	psw			; save registers
	push	acc
	push	0
	push	b
	push	c

	set1	psw,rambk0	; game ram bank
	ld	p3		; p3 state
	push	acc		; save p3 state
	xor	keys		; keys which changed status are now 1
	pop	keys		; get p3 state and store for next call
	bz	.nokeychange	; nothing changed, finished
	mov	#1,b		; number of key
	mov	#1,c		; bitmask of key
.keyloop
	push	acc		; store keymask
	and	c		; mask out one bit
	bz	.nokeyfound	; key is not pressed
	ld	keys		; get p3 state
	and	c		; mask out one bit
	bz	.keynorelease	; skip release marker
	mov	#$80,acc	; set release flag
.keynorelease
	or	b		; keynumber + release flag
	push	keyqin		; move queue input
	pop	0		; to pointer register
	st	@r0		; store key event in pointer
	st	lastkeyevent	; store for repeat
	; advance queue pointer
	ld	keyqin		; queue input pointer
	inc	acc		; advance
	bne	#keyqend+1,.keyqueueok1	; test for end of queue
	mov	#keyqstart,acc	; end of queue, start again
.keyqueueok1
	st	keyqin		; new queue input pointer
	; reset repeat counter
	ld	keyrepeat
	and	#$80		; mask out repeat flag, set rest to 0
	st	keyrepeat
	
.nokeyfound
	inc	b		; next keynumber
	ld	c
	clr1	psw,cy		; new bit to shift in is 0
	rolc			; next bit to test
	st	c
	pop	acc		; get keymask of all keys
	bn	psw,cy,.keyloop	; repeat until all bits tested

.nokeychange
	; test for repeat
	bn	keyrepeat,7,.timer0end
	inc	keyrepeat		; count interrupts until repeat
	ld	keyrepeat
	bne	#$8f,.timer0end		; 1 char every 15 ints
	mov	#$80,keyrepeat		; restart
	; do repeat
	ld	lastkeyevent		; latest key event
	bp	acc,7,.timer0end	; don't repeat key release event
	be	#KEY_SLEEP,.timer0end	; no repeat on SLEEP
	be	#KEY_MODE,.timer0end	; no repeat on MODE
	be	#KEY_A,.timer0end	; no repeat on A
	be	#KEY_B,.timer0end	; no repeat on B
	; insert into queue
	push	keyqin			; move queue input
	pop	0			; to pointer register
	st	@r0			; store key event in pointer
	; advance queue pointer
	ld	keyqin			; queue input pointer
	inc	acc			; advance
	bne	#keyqend+1,.keyqueueok2	; test for end of queue
	mov	#keyqstart,acc		; end of queue, start again
.keyqueueok2
	st	keyqin			; new queue input pointer

.timer0end
	pop	c			; restore registers
	pop	b
	pop	0
	pop	acc
	pop	psw
.timer0ret
	reti

; returns one pressed keynumber or 0 in acc
; changes keyboard queue pointers
getonekey
	clr1	t0con,t0hie	; disable keyboard irq
	ld	keyqin		; get input pointer
	xor	keyqout		; compare with output pointer
	bz	.getonekeyend	; no event in queue
	push	0		; save pointer register
	push	keyqout		; move queue output
	pop	0		; to pointer register
	ld	@r0		; store key event in pointer
	pop	0		; restore pointer register
	push	acc		; save key event
	; advance queue pointer
	ld	keyqout		; queue output pointer
	inc	acc		; advance
	bne	#keyqend+1,.getonequeueok	; test for end of queue
	mov	#keyqstart,acc	; end of queue, start again
.getonequeueok
	st	keyqout
	pop	acc		; restore key event
.getonekeyend
	set1	t0con,t0hie	; restore keyboard irq state
	ret

; return one pressed keynumber or 0 in acc
; handles sleep, mode, docked
; changes keyboard queue pointers, vccr
getkey
	call	getonekey
	bp	p7,0,.quit		; Docked?
	be	#KEY_MODE,.quit		; MODE pressed, quit
	be	#KEY_SLEEP,.sleep	; SLEEP pressed, go to sleep
	ret
.quit
	jmp	goodbye
.sleep
	mov	#0,vccr			; Blank LCD
.sleepmore:
	set1	pcon,HALT		; Enter HALT mode
	call	getonekey
	bp	p7,0,.quit		; Docked?
	bne	#KEY_SLEEP,.sleepmore	; No SLEEP press yet
	; wake up again
	mov	#$80,vccr		; Reenable LCD
	br	getkey			; check again for keys

; clear the keyboard queue
; destroys acc
clearkey
	clr1	t0con,t0hie		; disable repeat irq
	ld	keyqin			; get input pointer
	st	keyqout			; set to output pointer
	mov	#$80,lastkeyevent	; clear key event
	set1	t0con,t0hie		; disable repeat irq
	ret

; --- keyboard code end here ---

; --- sound engine code starts here ---
; available as seperate package at http://soeren.informationstheater.de/vm/sound.html

; start to play tune pointed by trh/trl
; destroys acc
playtune
	mov	#0,t1cnt	; stop timer
	ld	trl		; copy pointer to tune
	st	soundpointerl
	ld	trh
	st	soundpointerh
	mov	#$ff,t1hr	; timer rolls over immediately
	mov	#0,soundwait	; next event immediately
	mov	#$ff,soundplaying	; we are playing
	mov	#$94,t1cnt	; start timerh, irq, rest is done in irq
	ret

; handle irq requests related to playing a tune
sound_irq
	bn	t1cnt,t1hovf,.sound_iend	; skip if not timer1 high ovl
	clr1	t1cnt,t1hovf	; clear overflow, important!
	push	psw		; store registers
	push	acc
	set1	psw,rambk0	; game ram bank
	ld	soundwait	; need to wait?
	bz	.soundevent	; no, next event
	dec	soundwait	; we have completed one wait cycle
	pop	acc		; restore registers
	pop	psw
.sound_iend
	reti
	; process sound event
.soundevent
	push	trl		; store registers
	push	trh
.soundeventloop
	ld	soundpointerl	; get pointer to next sound event
	st	trl
	ld	soundpointerh
	st	trh
	mov	#0,acc		; offset for pointer
	ldc			; command byte of sound event
	be	#1,.soundnote	; 1 = set note
	be	#2,.soundwait8	; 2 = wait
	; unknown command or 0, stop sound
	mov	#0,soundplaying	; set flag
	mov	#0,t1cnt	; stop timer
	pop	trh		; restore registers
	pop	trl
	pop	acc
	pop	psw
	reti
	; play note
.soundnote
	clr1	t1cnt,t1lrun	; stop timer1 low (sound)
	mov	#1,acc		; offset to value for t1lr
	ldc
	st	t1lr
	mov	#2,acc		; offset to value for t1lc
	ldc
	st	t1lc
	set1	t1cnt,t1lrun	; start timer1 low (sound)
	set1	t1cnt,eldt1c	; set compare value
	clr1	t1cnt,t1long	; use 2 8 bit timers
	; advance soundpointer
	ld	soundpointerl	; low byte
	add	#3		; command was 3 bytes long
	st	soundpointerl
	bn	psw,cy,.sound1	; skip next unless overflow
	inc	soundpointerh	; advance high byte
.sound1
	br	.soundeventloop	; note is started, next event
	; wait 8bit
.soundwait8
	mov	#1,acc		; offset to number of wait cycles
	ldc
	st	soundwait
	; advance soundpointer
	ld	soundpointerl	; low byte
	add	#2		; command was 2 bytes long
	st	soundpointerl
	bn	psw,cy,.sound2	; skip next unless overflow
	inc	soundpointerh	; advance high byte
.sound2
	mov	#$00,t1hr	; timer1 high period to maximum
	set1	t1cnt,t1hrun	; start timer1 high (wait)
	set1	t1cnt,t1hie	; start timer1 irq (wait)
	pop	trh		; restore registers
	pop	trl
	pop	acc
	pop	psw
	reti

; --- sound engine code ends here ---

; the font for the playfield, 5x5 pixel used on normal chars, 6x5 pixel drawn
font5x5:
	; 0
	.byte	%01110000
	.byte	%10011000
	.byte	%10101000
	.byte	%11001000
	.byte	%01110000
	; 1
	.byte	%00100000
	.byte	%01100000
	.byte	%00100000
	.byte	%00100000
	.byte	%01110000
	; 2
	.byte	%01110000
	.byte	%10001000
	.byte	%00110000
	.byte	%01000000
	.byte	%11111000
	; 3
	.byte	%01110000
	.byte	%10001000
	.byte	%00110000
	.byte	%10001000
	.byte	%01110000
	; 4
	.byte	%00110000
	.byte	%01010000
	.byte	%10010000
	.byte	%11111000
	.byte	%00010000
	; 5
	.byte	%11111000
	.byte	%10000000
	.byte	%11110000
	.byte	%00001000
	.byte	%11110000
	; 6
	.byte	%01110000
	.byte	%10000000
	.byte	%11110000
	.byte	%10001000
	.byte	%01110000
	; 7
	.byte	%11111000
	.byte	%00001000
	.byte	%00010000
	.byte	%00100000
	.byte	%00100000
	; 8
	.byte	%01110000
	.byte	%10001000
	.byte	%01110000
	.byte	%10001000
	.byte	%01110000
	; 9
	.byte	%01110000
	.byte	%10001000
	.byte	%01111000
	.byte	%00001000
	.byte	%01110000
	; ? = 10
	.byte	%01110000
	.byte	%10001000
	.byte	%00110000
	.byte	%00000000
	.byte	%00100000
	; bomb = 11
	.byte	%00010000
	.byte	%00100000
	.byte	%01110000
	.byte	%11111000
	.byte	%01110000
	; flag = 12
	.byte	%01111000
	.byte	%01111000
	.byte	%01000000
	.byte	%01000000
	.byte	%11100000
	; cursor = 13
	.byte	%00000000
	.byte	%01010000
	.byte	%00100000
	.byte	%01010000
	.byte	%00000000
	; left of you = 14
	.byte	%10100100
	.byte	%10101000
	.byte	%10101000
	.byte	%01001000
	.byte	%01000100
	; right of you = 15
	.byte	%10010100
	.byte	%01010100
	.byte	%01010100
	.byte	%01010100
	.byte	%10011100
	; left of win = 16
	.byte	%10001000
	.byte	%10001000
	.byte	%10001000
	.byte	%10101000
	.byte	%01010000
	; right of win = 17
	.byte	%10100100
	.byte	%10110100
	.byte	%10111100
	.byte	%10101100
	.byte	%10100100
	; shifted O = 18
	.byte	%00111000
	.byte	%01000100
	.byte	%01000100
	.byte	%01000100
	.byte	%00111000
	; shifted V = 19
	.byte	%01000100
	.byte	%01000100
	.byte	%01000100
	.byte	%00101000
	.byte	%00010000
	; shifted E = 20
	.byte	%01111100
	.byte	%01000000
	.byte	%01110000
	.byte	%01000000
	.byte	%01111100
	; shifted R = 21
	.byte	%01111000
	.byte	%01000100
	.byte	%01111000
	.byte	%01001000
	.byte	%01000100
	; [SPACE] = 22
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	; A = 23
	.byte	%01110000
	.byte	%10001000
	.byte	%11111000
	.byte	%10001000
	.byte	%10001000
	; B = 24
	.byte	%11110000
	.byte	%10001000
	.byte	%11110000
	.byte	%10001000
	.byte	%11110000
	; C = 25
	.byte	%01111000
	.byte	%10000000
	.byte	%10000000
	.byte	%10000000
	.byte	%01111000
	; D = 26
	.byte	%11110000
	.byte	%10001000
	.byte	%10001000
	.byte	%10001000
	.byte	%11110000
	; E = 27
	.byte	%11111000
	.byte	%10000000
	.byte	%11100000
	.byte	%10000000
	.byte	%11111000
	; F = 28
	.byte	%11111000
	.byte	%10000000
	.byte	%11100000
	.byte	%10000000
	.byte	%10000000
	; G = 29
	.byte	%01111000
	.byte	%10000000
	.byte	%10111000
	.byte	%10001000
	.byte	%01110000
	; H = 30
	.byte	%10001000
	.byte	%10001000
	.byte	%11111000
	.byte	%10001000
	.byte	%10001000
	; I = 31
	.byte	%11111000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000
	.byte	%11111000
	; J = 32
	.byte	%11111000
	.byte	%00001000
	.byte	%00001000
	.byte	%10001000
	.byte	%01110000
	; K = 33
	.byte	%10001000
	.byte	%10010000
	.byte	%11100000
	.byte	%10010000
	.byte	%10001000
	; L = 34
	.byte	%10000000
	.byte	%10000000
	.byte	%10000000
	.byte	%10000000
	.byte	%11111000
	; M = 35
	.byte	%10001000
	.byte	%11011000
	.byte	%10101000
	.byte	%10001000
	.byte	%10001000
	; N = 36
	.byte	%10001000
	.byte	%11001000
	.byte	%10101000
	.byte	%10011000
	.byte	%10001000
	; O = 37
	.byte	%01110000
	.byte	%10001000
	.byte	%10001000
	.byte	%10001000
	.byte	%01110000
	; P = 38
	.byte	%11110000
	.byte	%10001000
	.byte	%11110000
	.byte	%10000000
	.byte	%10000000
	; Q = 39
	.byte	%01110000
	.byte	%10001000
	.byte	%10101000
	.byte	%10010000
	.byte	%01101000
	; R = 40
	.byte	%11110000
	.byte	%10001000
	.byte	%11110000
	.byte	%10010000
	.byte	%10001000
	; S = 41
	.byte	%01111000
	.byte	%10000000
	.byte	%01110000
	.byte	%00001000
	.byte	%11110000
	; T = 42
	.byte	%11111000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000
	; U = 43
	.byte	%10001000
	.byte	%10001000
	.byte	%10001000
	.byte	%10001000
	.byte	%01110000
	; V = 44
	.byte	%10001000
	.byte	%10001000
	.byte	%10001000
	.byte	%01010000
	.byte	%00100000
	; W = 45
	.byte	%10001000
	.byte	%10001000
	.byte	%10001000
	.byte	%10101000
	.byte	%01010000
	; X = 46
	.byte	%10001000
	.byte	%01010000
	.byte	%00100000
	.byte	%01010000
	.byte	%10001000
	; Y = 47
	.byte	%10001000
	.byte	%01010000
	.byte	%00100000
	.byte	%00100000
	.byte	%00100000
	; Z = 48
	.byte	%11111000
	.byte	%00010000
	.byte	%00100000
	.byte	%01000000
	.byte	%11111000
	; . = 49
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%01110000
	.byte	%01110000

; the font for the time display, 3x5 pixel used on normal chars, 3x5 pixel drawn
font3x5:
	; 0
	.byte	%11100000
	.byte	%10100000
	.byte	%10100000
	.byte	%10100000
	.byte	%11100000
	; 1
	.byte	%01000000
	.byte	%11000000
	.byte	%01000000
	.byte	%01000000
	.byte	%11100000
	; 2
	.byte	%11100000
	.byte	%00100000
	.byte	%11100000
	.byte	%10000000
	.byte	%11100000
	; 3
	.byte	%11100000
	.byte	%00100000
	.byte	%01100000
	.byte	%00100000
	.byte	%11100000
	; 4
	.byte	%10000000
	.byte	%10000000
	.byte	%11100000
	.byte	%01000000
	.byte	%01000000
	; 5
	.byte	%11100000
	.byte	%10000000
	.byte	%11100000
	.byte	%00100000
	.byte	%11100000
	; 6
	.byte	%11100000
	.byte	%10000000
	.byte	%11100000
	.byte	%10100000
	.byte	%11100000
	; 7
	.byte	%11100000
	.byte	%00100000
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000
	; 8
	.byte	%11100000
	.byte	%10100000
	.byte	%11100000
	.byte	%10100000
	.byte	%11100000
	; 9
	.byte	%11100000
	.byte	%10100000
	.byte	%11100000
	.byte	%00100000
	.byte	%11100000

; Titlescreen
titlebitmap
	.include "title.i"

; Show highscores screen
showhighbitmap
	.include "highscore.i"

; Select # of bombs screen
bombbitmap
	.include "bombs.i"

; New highscore screen
newhighbitmap
	.include "newhigh.i"

; Flash error screen
flasherrbitmap
	.include "flasherr.i"

; Tune to play at winning games
wintune
	.byte $01,$ec,$f6,$02,$02
	.byte $01,$ee,$f7,$02,$02
	.byte $01,$f0,$f8,$02,$08
	.byte $01,$f6,$fb,$02,$02
	.byte $01,$ff,$fe,$02,$01
	.byte $01,$f6,$fb,$02,$02
	.byte 0

; Tune to play at highscores
hightune
	.byte $01,$ec,$f6,$02,$04
	.byte $01,$ee,$f7,$02,$04
	.byte $01,$f0,$f8,$02,$08
	.byte $01,$f1,$f8,$02,$04
	.byte $01,$f3,$f9,$02,$08
	.byte $01,$f0,$f8,$02,$04
	.byte $01,$ec,$f6,$02,$08
	.byte $01,$f0,$f8,$02,$04
	.byte $01,$f3,$f9,$02,$04
	.byte $01,$f6,$fb,$02,$10
	.byte 0

; Tune to play at explosions
explodetune
	.byte $01,$f3,$f9,$02,$02
	.byte $01,$f0,$f8,$02,$02
	.byte $01,$ec,$f6,$02,$08
	.byte 0

; binary version tag
	.byte	13,10,"$Id: minesweeper.s,v 1.31 2003/07/19 13:18:17 sgust Exp $",13,10,0

; make sure that the highscores start at block boundary for easier flashing
	.cnop	0,$80

; initial highscore table
highscores
	.byte	$09,$99,41,49,29,0,0,0		; 2 bombs
	.byte	$09,$99,41,49,29,0,0,0		; 3 bombs
	.byte	$09,$99,41,49,29,0,0,0		; 4 bombs
	.byte	$09,$99,41,49,29,0,0,0		; 5 bombs
	.byte	$09,$99,41,49,29,0,0,0		; 6 bombs
	.byte	$09,$99,41,49,29,0,0,0		; 7 bombs
	.byte	$09,$99,41,49,29,0,0,0		; 8 bombs
	.byte	$09,$99,41,49,29,0,0,0		; 9 bombs

	.cnop	0,$200		; pad to an even number of blocks
end
